package agent;

import queue.*;
import java.util.*;

public class ForwardCheckingSearch extends SearchMethod {

    //Como Quick Search, o algoritmo utiliza as heuristicas:
    //most constrained variable e least constrainig value
    public static int QUICK_SEARCH = 0;
    
    //Como Normal Search o algoritmo utiliza somente a heuristica
    //most constrained variable
    public static int NORMAL_SEARCH = 1;
    
    //Tipo da busca, pode ser Quick Search ou Normal Search
    private int searchType;

    //Tamanho do problema
    private int n;
    
    //Domínios das variáveis
    private boolean[][] domains;
    
    //Ranking de valores afetados;
    private int[] ranking;
    
    public ForwardCheckingSearch(int n) {
        setN(n);
        this.searchType = NORMAL_SEARCH;
    }

    public ForwardCheckingSearch(int n, int searchType) {
        setN(n);
        setSearchType(searchType);
    }
    
    protected Collection expand(Node node) {
        generateDomains(node.getState().getVariables());
        int mostConstrainedVariable = getMostConstrainedVariable();
        if (mostConstrainedVariable >= 0) {
            int[] values = getValues(node.getState(), mostConstrainedVariable);
            ArrayList nodes = new ArrayList(values.length);
            State newState = null;
            for (int i=0; i<values.length; i++) {
                if (values[i] != 0) {
                    newState = new State();
                    newState.setN(node.getState().getN());
                    newState.setVariables(node.getState().getVariables());
                    assign(newState, mostConstrainedVariable, values[i]);
                    nodes.add(0, new Node(newState));
                }
            }
            return nodes;
        }
        return null;
    }
    
    private void assign(State state, int variable, int value) {
        int[] variables = state.getVariables();
        variables[variable] = value;
    }
    
    private int getMostConstrainedVariable() {
        int mcv = 0;
        for (int i=0; i<ranking.length; i++) {
            if ( (ranking[mcv] == n) ||
                 (ranking[i] < n && ranking[i] > ranking[mcv]) ) {
                mcv = i;
            }
        }
        if (ranking[mcv] == n) {
            return -1;
        }
        return mcv;
    }
    
    private int[] getValues(State state, int variable) {
        if (searchType == QUICK_SEARCH) {
            return getLeastConstrainingValues(state, variable);
        } else if (variable >= 0 && variable <= n) {
            int[] values = new int[n];
            int j = 0;
            int nav = n + 1;
            for (int i=0; i<domains[variable].length; i++) {
                if (!domains[variable][i]) {
                    values[j++] = i+1;
                }
            }
            return values;
        }
        return null;
    }
    
    private int[] getLeastConstrainingValues(State state, int variable) {
        if (variable >= 0 && variable < n) {
            int[] lcv = new int[n];
            int j = 0;
            int nav = 4*n;
            for (int i=0; i<domains[variable].length; i++) {
                int aux = nav;
                if (!domains[variable][i]) {
                    aux = getNumberOfAfectedValues(state, variable, i);
                    if (aux < nav) {
                        nav = aux;
                        lcv = new int[n];
                        lcv[0] = i+1;
                        j = 1;
                    } else if (aux == nav){
                        lcv[j++] = i+1;
                    }
                }
            }
            return lcv;
        }
        return null;
    }
    
    private int getNumberOfAfectedValues(State state, int variable, int value) {
        int[] variables = state.getVariables();
        
        int nav = 0;
        for (int i=0; i<domains.length; i++) {
            
            //verifica linha
            if (!domains[variable][i]) {
                nav++;
            }
            
            if (i!=variable && variables[i]==0) {
                
                //verifica coluna
                if (!domains[i][value]) {
                    nav++;
                }
                int diferenca = variable - i;
                
                //verifica diagonais
                if ( (value-diferenca >= 0) && (value-diferenca < n) && (!domains[i][value-diferenca])) {
                    nav++;
                }
                if ( (value+diferenca >= 0) && (value+diferenca < n) && (!domains[i][value+diferenca])) {
                    nav++;
                }
            }
        }
        return nav;
    }
    
    private void generateDomains(int[] variables) {
        ranking = new int[variables.length];
        domains = new boolean[variables.length][variables.length];
        
        for (int i=0; i<variables.length; i++) {
            if (variables[i] != 0) {
                updateDomains(variables, i, variables[i]);
            }
        }
    }
    
    private void updateDomains(int[] variables, int variable, int value) {
        
        value = value - 1;
        
        for (int i=0; i<domains.length; i++) {
            
            //corta linha
            removeFromDomains(variable, i);
            
            if (i!=variable && variables[i]==0) {
                
                //corta coluna
                removeFromDomains(i, value);
                
                int diferenca = variable - i;
                
                //corta diagonais
                if ( (value-diferenca >= 0) && (value-diferenca < variables.length)) {
                    removeFromDomains(i, value-diferenca);
                }
                if ( (value+diferenca >= 0) && (value+diferenca < variables.length)) {
                    removeFromDomains(i, value+diferenca);
                }
            }
        }
    }
    
    private void removeFromDomains(int variable, int value) {
        if (!domains[variable][value]) {
            domains[variable][value] = true;
            ranking[variable]++;
        }
    }
    
    /** Getter for property searchType.
     * @return Value of property searchType.
     */
    public int getSearchType() {
        return searchType;
    }
    
    /** Setter for property searchType.
     * @param searchType New value of property searchType.
     */
    public void setSearchType(int searchType) {
        this.searchType = searchType;
    }
    
    /** Getter for property n.
     * @return Value of property n.
     */
    public int getN() {
        return n;
    }
    
    /** Setter for property n.
     * @param n New value of property n.
     */
    public void setN(int n) {
        this.n = n;
    }
}
